#ifndef HARKIO_DEFINEITEM_HPP_
#define HARKIO_DEFINEITEM_HPP_

#include <harkio2/NotInitialized.hpp>
#include <harkio2/InitializedFlag.hpp>

/** @file defineitem.hpp
 * @brief Macros to define an item in data structure.
 */

/** @brief Defines a flag variable for the item state (initialized or not).
 * @param item_name The name of item
 */
#define DEFINE_INITIALIZED_FLAG(item_name) \
    private: InitializedFlag item_name ## _is_initialized;

/** @brief Defines a getter for the flag variable.
 * @param item_name The name of this item
 * @param accessor_name The name to be used for creating this getter name
 */
#define DEFINE_ISINITIALIZED(item_name, accessor_name) \
    public: InitializedFlag Is ## accessor_name ## Initialized () const { \
        return item_name ## _is_initialized; \
    }

/** @brief Defines a item with the given name 'item_name'.
 * @param item_name The name of item
 * @prram item_type The type of the item
 */
#define DEFINE_ITEM_VARIABLE(item_name, item_type) \
    private: item_type item_name;


/** @brief Defines a getter for the item.
 * This getter:
 *   -when the item has a value -> returns the value
 *   -otherwise (= the item hasn't been initialized yet) ->
 *     throws NotInitialized exception
 * @param item_name The name of item
 * @param accessor_name The name of this getter (specify XXX in GetXXX)
 * @param accessor_type The return type of this getter
 * @param access The access specifier for this getter.
 */
#define DEFINE_GETTER(item_name, accessor_name, accessor_type, access) \
    access: accessor_type Get ## accessor_name () const { \
        if (item_name ## _is_initialized) { return item_name; } \
        else { throw NotInitialized( #item_name " is not initialized" ); } \
    }

/** @brief Defines a setter for the item.
 * This setter sets a given value to the item and sets a flag as initialized.
 * @param item_name The name of item
 * @param accessor_name The name of this setter (specify XXX in SetXXX)
 * @param accessor_type The argument type of this setter
 * @param access The access specifier for this setter.
 */
#define DEFINE_SETTER(item_name, accessor_name, accessor_type, access) \
    access: void Set ## accessor_name (accessor_type value) { \
        item_name = value; \
        item_name ## _is_initialized = true; \
    }

/** @brief Defines an item, a getter and a setter.
 * Defines a variable for the item and defines a getter and a setter.
 * The item will be readable and writable.
 * @param item_name The name of item
 * @prram item_type The type of the item
 * @param accessor_name The name of accessors (specity XXX in GetXXX and SetXXX)
 * @param accessor_type The return and argument type of accessors
 */
#define DEFINE_ITEM(item_name, item_type, accessor_name, accessor_type) \
    DEFINE_ITEM_VARIABLE(item_name, item_type) \
    DEFINE_INITIALIZED_FLAG(item_name) \
    DEFINE_ISINITIALIZED(item_name, accessor_name) \
    DEFINE_GETTER(item_name, accessor_name, accessor_type, public) \
    DEFINE_SETTER(item_name, accessor_name, accessor_type, public)

/** @brief Defines an item and a getter.
 * Defines a variable for the item and defines a getter.
 * The item will be read only, so the item can't be read
 * unless another method sets the flag variable as initialised
 * (writing true to the flag or using private setter).
 * @param item_name The name of item
 * @prram item_type The type of the item
 * @param accessor_name The name of the getter (specity XXX in GetXXX)
 * @param accessor_type The return type of the getter
 */
#define DEFINE_ITEM_READONLY(item_name, item_type, accessor_name, accessor_type) \
    DEFINE_ITEM_VARIABLE(item_name, item_type) \
    DEFINE_INITIALIZED_FLAG(item_name) \
    DEFINE_ISINITIALIZED(item_name, accessor_name) \
    DEFINE_GETTER(item_name, accessor_name, accessor_type, public) \
    DEFINE_SETTER(item_name, accessor_name, accessor_type, private)

#endif // HARKIO_DEFINEITEM_HPP_

